/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "MidiFileReader.h"
#include "Midi/Sequence.h"
#include "Midi/Track.h"
#include "GUI/GraphicalTrack.h"
#include "Pickers/InstrumentChoice.h"
#include "Pickers/DrumChoice.h"
#include "GUI/MainFrame.h"
#include "Dialogs/WaitWindow.h"
#include "Editors/DrumEditor.h"
#include "Editors/KeyboardEditor.h"

#include "jdkmidi/world.h"
#include "jdkmidi/track.h"
#include "jdkmidi/multitrack.h"
#include "jdkmidi/filereadmultitrack.h"
#include "jdkmidi/fileread.h"
#include "jdkmidi/fileshow.h"
#include "jdkmidi/filewritemultitrack.h"

#include <cmath>
#include <string>

namespace AriaMaestosa {

bool loadMidiFile(Sequence* sequence, char* filepath)
{
    
    sequence->importing = true;
    
    // the stream used to read the input file
    jdkmidi::MIDIFileReadStreamFile rs( filepath );
    
    // the object which will hold all the tracks
    jdkmidi::MIDIMultiTrack jdksequence;
    
    // the object which loads the tracks into the tracks object
    jdkmidi::MIDIFileReadMultiTrack track_loader( &jdksequence );
    
    // the object which parses the midifile and gives it to the multitrack loader
    jdkmidi::MIDIFileRead reader( &rs, &track_loader );
    
    // load the midifile into the multitrack object
    if( !reader.Parse() )
	{
		std::cout << "Error: could not parse midi file" << std::endl;
		return false;
	}
    
    /*
     MIDITrack MIDIMultiTrack::*GetTrack( int trk );
     const MIDITrack MIDIMultiTrack::*GetTrack( int trk ) const;
     int MIDIMultiTrack::GetNumTracks()
     
     ----------------------------------------------------------------------------------------------------
     class  MIDITrack 	
     ----------------------------------------------------------------------------------------------------

         *** const 	MIDITimedBigMessage *GetEvent( int event_num ) const;
         *** int	GetNumEvents() const;

         
         MIDITrack( int size=0 );
         
         MIDITrack( const MIDITrack &t );
         
         ~MIDITrack();
         
         void	Clear();
         void	Shrink();
         
         void	ClearAndMerge( const MIDITrack *src1, const MIDITrack *src2 );

         bool	Expand( int increase_amount=(MIDITrackChunkSize) );
         
         MIDITimedBigMessage * GetEventAddress( int event_num );
         
         const MIDITimedBigMessage * GetEventAddress( int event_num ) const;
         
         const 	MIDITimedBigMessage *GetEvent( int event_num ) const;
         MIDITimedBigMessage *GetEvent( int event_num );
         bool	GetEvent( int event_num, MIDITimedBigMessage *msg ) const;
         
         bool	PutEvent( const MIDITimedBigMessage &msg );
         bool	PutEvent( const MIDITimedMessage &msg, MIDISystemExclusive *sysex );
         bool	SetEvent( int event_num, const MIDITimedBigMessage &msg );
         
         bool	MakeEventNoOp( int event_num );
         
         bool	FindEventNumber( MIDIClockTime time, int *event_num ) const;
         
         int	GetBufferSize()	const;
         int	GetNumEvents() const;

         }
     
     ----------------------------------------------------------------------------------------------------
     MIDITIMEDBIGMESSAGE:
     ----------------------------------------------------------------------------------------------------

     MIDIClockTime	GetTime() const;
      MIDISystemExclusive *GetSysEx();
     
     char	GetLength() const;
     
     unsigned char	GetStatus() const	{ return (unsigned char)status;	}
     unsigned char	GetChannel() const	{ return (unsigned char)(status&0x0f);	}
     unsigned char	GetType() const		{ return (unsigned char)(status&0xf0);	}
     
     unsigned char	GetMetaType() const	{ return byte1;		}
     
     unsigned char	GetByte1() const	{ return byte1;		}
     unsigned char	GetByte2() const	{ return byte2;		}
     unsigned char	GetByte3() const	{ return byte3;		}
     
     unsigned char	GetNote() const		{ return byte1;		}
     unsigned char	GetVelocity() const	{ return byte2;		}
     unsigned char	GetPGValue() const	{ return byte1;		}
     unsigned char	GetController() const	{ return byte1;		}
     unsigned char	GetControllerValue() const { return byte2;	}
     
     short	GetBenderValue() const;
     
     unsigned short	GetMetaValue()	const;
     unsigned char GetTimeSigNumerator() const;
     unsigned char GetTimeSigDenominator() const;
     signed char GetKeySigSharpFlats() const;
     unsigned char GetKeySigMajorMinor() const;
     
     bool	IsChannelMsg() const;
     
     bool	IsNoteOn() const;
     bool	IsNoteOff() const;
     bool	IsPolyPressure() const;
     bool	IsControlChange() const;
     bool	IsProgramChange() const;
     bool	IsChannelPressure() const;
     bool	IsPitchBend() const;
     bool	IsSystemMessage() const;
     bool	IsSysEx() const;
     short	GetSysExNum() const;
     bool	IsMTC() const;
     bool	IsSongPosition() const;
    bool	IsSongSelect() const;
     bool 	IsTuneRequest() const;
     bool	IsMetaEvent() const; 
     bool 	IsTextEvent() const;
     bool	IsAllNotesOff() const;
     bool	IsNoOp() const;
     bool	IsTempo() const;
     bool	IsDataEnd() const;
     bool 	IsTimeSig() const;
     bool      IsKeySig() const;
     bool 	IsBeatMarker() const;
     
     //
     // GetTempo() returns the tempo value in 1/32 bpm
     //
     
     unsigned short	GetTempo32() const;
     
     unsigned short	GetLoopNumber() const;
     
     */
    
    jdkmidi::MIDITrack* track;
    jdkmidi::MIDITimedBigMessage* event;
    
    int resolution = jdksequence.GetClksPerBeat();
    sequence->setTicksPerBeat( resolution );
        
    sequence->prepareEmptyTracksForLoading(16);
    
    int lastEventTick = 0;
    
    bool firstTempoEvent=true;
    
    const int trackAmount = jdksequence.GetNumTracks();
    
     // ----------------------------------- for each track -------------------------------------
    for(int trackID=0; trackID<trackAmount; trackID++)
	{
        track = jdksequence.GetTrack( trackID );
        
        WaitWindow::setProgress(
                                (int)(
                                      log(trackID)*100/log(trackAmount)
                                      )
                                );
        wxYield();
        
         // ----------------------------------- for each event -------------------------------------
        
        int last_channel = -1;
        std::string trackName = wxGetTranslation("Untitled");
        
		const int eventAmount = track->GetNumEvents();
		// FIXME - only true in one track per channel cases
		//sequence->getTrack(trackID)->reserveMemoryForEvents(eventAmount);

        for(int eventID=0; eventID<eventAmount; eventID++)
		{
            
            event = track->GetEvent( eventID );
            
            const int tick = event->GetTime();
            if(tick > lastEventTick) lastEventTick = tick;
            
            const int channel = event->GetChannel();
            
            
            if(!event->IsMetaEvent())
			{
                // check if all notes in one track all belong to the same channel, and find the channel that is mainly used (for track names, etc.)
                if( channel != last_channel and last_channel != -1)
				{
                    static bool already_shown = false; // only print message one
					
					if(!already_shown)
					{
						std::cout << "/!\\ WARNING: midi is not 'one track-one channel' formatted. VV may not handle this properly. (" <<
						last_channel << ", " << channel << ")" << std::endl; 
						already_shown=true;
                    }
                    last_channel = channel;
                    
                }
				else
				{
                    last_channel = channel;
                }
                
            }
           
            // ----------------------------------- note on -------------------------------------
            if( event->IsNoteOn() )
			{
                
                const int note = ((channel == 9) ? event->GetNote() : 131 - event->GetNote());
                const int volume = event->GetVelocity();
                
                if(!sequence->getTrack(channel)->addNote(note,
														 tick,
														 tick+1 /*temporary end until the corresponding note off event is found*/,
														 volume))
				{
                    std::cout << "A note could not be added" << std::endl;
                    //sequence->prepareEmptyTracksForLoading(1); // clear any partial load
                    //return false;
                }
                
            }
			// ----------------------------------- note off -------------------------------------
			else if( event->IsNoteOff() )
			{
				
                const int note = ((channel == 9) ? event->GetNote() : 131 - event->GetNote());

                // a note off event was found, find to which note on event it corresponds
                // (start iterating from the end, because since events are in order it will probably be found near the end)
                for(int n=sequence->getTrack(channel)->getNoteAmount()-1; n>-1; n--)
				{
                    
                    if(sequence->getTrack(channel)->getNotePitchID(n) == note and
                       sequence->getTrack(channel)->getNoteEndInMidiTicks(n)==sequence->getTrack(channel)->getNoteStartInMidiTicks(n)+1 )
					{
                        sequence->getTrack(channel)->setNoteEnd( tick, n );
                        
                        assertExpr(sequence->getTrack(channel)->getNoteEndInMidiTicks(n), ==, tick);
                        
                        break;
                    }//end if
                    
                    
                    if(n==0)
					{
                        std::cout << "Failed to link note off to a corresponding note on event. (ignoring it)" << " // tick: " << tick << ", note: " << note << ", channel: " << channel << std::endl;
                    }
                } // next
                
            }
			// ----------------------------------- control change -------------------------------------
			else if( event->IsControlChange() )
			{
                const int controllerID = event->GetController();
                const int value = 127-event->GetControllerValue();
                    
                sequence->getTrack(channel)->addControllerFromMidiFileReader(tick, value, controllerID);
               
            }
			// ----------------------------------- pitch bend -------------------------------------
			else if( event->IsPitchBend() )
			{
                
                int pitchBendVal = event->GetBenderValue();
                
                int value = (int)round( (pitchBendVal+8064.0)*128.0/16128.0 );
                if(value>127) value=127;
                if(value<0) value=0;
                
                sequence->getTrack(channel)->addControllerFromMidiFileReader(tick, 127-value, 200);
                
            }
			// ----------------------------------- program chnage -------------------------------------
			else if( event->IsProgramChange() )
			{
				
                const int instrument = event->GetPGValue();
                
                if(channel == 9)  sequence->getTrack(channel)->graphics->drumEditor->drumKit->setDrumID(instrument);
                else			  sequence->getTrack(channel)->graphics->keyboardEditor->instrument->setInstrumentID(instrument);
                
            }
			// ----------------------------------- tempo -------------------------------------
			else if( event->IsTempo() )
			{
                
                const int tempo = event->GetTempo32()/32;
                
                if(firstTempoEvent)
				{
                    
                    sequence->setTempo(tempo);
                    
                    char buffer[4];
                    sprintf (buffer, "%d", tempo);
                    
                    sequence->mainFrame->tempoCtrl->SetValue(buffer);
                    firstTempoEvent=false;
                    
                }
				else
				{
                    
                    sequence->getTrack(0)->addController(tick, /*FIXME - clarify this*/127-(int)((tempo-20.0)/380.0*128.0) , 201);
					
                }
                // ----------------------------------- time key/sig and beat marker -------------------------------------
            }
			else if( event->IsTimeSig() )
			{
				
				sequence->setMeasureSig( (int)event->GetTimeSigNumerator(), (int)event->GetTimeSigDenominator() );
				
            }
			/*
			else if( event->IsKeySig() )
			{
                //std::cout << "Key sig: " << (int)event->GetKeySigSharpFlats() << " " << (int)event->GetKeySigMajorMinor() << std::endl;
            }
			else if( event->IsBeatMarker() )
			{
                //std::cout << "Beat marker: " << (int)event->GetByte1() << " " << (int)event->GetByte2() << " " << (int)event->GetByte3() << ", type=" << (int)event->GetType() << std::endl;
            }
			 */
			else  if( event->IsTextEvent() ) // ----------------------------------- track name / text events -------------------------------------
			{
				
				// sequence/track name
                if((int)event->GetByte1() == 3)
				{ 
                    
                    const int length = event->GetSysEx()->GetLength(); // get name length
                    char name[length+1];
                    name[length] = 0; // make zero-terminated
                    
                    char* buf = (char*) event->GetSysEx()->GetBuf();
                    for(int n=0; n<length; n++) name[n] = buf[n];
                    
                    trackName = name;

                }
				else if((int)event->GetByte1() == 2) // copyright
				{
                    
                    const int length = event->GetSysEx()->GetLength(); // get copyright length
                    char copyright[length+1];
                    copyright[length] = 0; // make zero-terminated
                    
                    char* buf = (char*) event->GetSysEx()->GetBuf();
                    for(int n=0; n<length; n++) copyright[n] = buf[n];
                    
                    sequence->setCopyright(copyright); 
                    
                }
				else
				{  
                    //const char* text = (char*) event->GetSysEx()->GetBuf();
                    //std::cout << "Text=" << text << ", channel="<< channel << ", byte1=" << (int)event->GetByte1() << std::endl;
                }
                
                
            }
			
			else{ std::cout << "ignored event" << std::endl; }
			
			/*
			else if( event->IsSystemMessage() )
			{
                std::cout << "System Message: " << (int)event->GetByte1() << " " << (int)event->GetByte2() << " " << (int)event->GetByte3() << " " << (int)event->GetType() << std::endl;
            }
			else if( event->IsSysEx() )
			{
                std::cout << "Sysex: " << (int)event->GetByte1() << " " << (int)event->GetByte2() << " " << (int)event->GetByte3() << " " << (int)event->GetType() << std::endl;
            }
			else if( event->IsMetaEvent() )
			{
                std::cout << "Meta Event: " << (int)event->GetByte1() << " " << (int)event->GetByte2() << " " << (int)event->GetByte3() << ", type= " << (int)event->GetType();
                
                if((int)event->GetByte1() == 0) std::cout << "---> sequence number" << std::endl;
                if((int)event->GetByte1() == 1) std::cout << "---> text" << std::endl;
                if((int)event->GetByte1() == 2) std::cout << "---> copyright" << std::endl;
                if((int)event->GetByte1() == 3) std::cout << "---> track name" << std::endl;
                if((int)event->GetByte1() == 4) std::cout << "---> instrument name" << std::endl;
                if((int)event->GetByte1() == 5) std::cout << "---> lyrics name" << std::endl;
                if((int)event->GetByte1() == 6) std::cout << "---> marker" << std::endl;
                if((int)event->GetByte1() == 7) std::cout << "---> cue point" << std::endl;
                
                if((int)event->GetByte1() == 32) std::cout << "---> prefix" << std::endl;
                
                if((int)event->GetByte1() == 47) std::cout << "---> end of track" << std::endl;
                
                if((int)event->GetByte1() == 81) std::cout << "---> tempo" << std::endl;
                
                if((int)event->GetByte1() == 84) std::cout << "---> SMPTE" << std::endl;
                if((int)event->GetByte1() == 88) std::cout << "---> Time signature" << std::endl;
                
                if((int)event->GetByte1() == 240) std::cout << "---> Sys Ex" << std::endl;
                
            }
			else
			{
                std::cout << "Unknown --> " << (char)event->GetByte1() << " " << (char)event->GetByte2() << " " << (char)event->GetByte3() << std::endl;
            }
*/
        }//next event
        
        if(last_channel != -1) sequence->getTrack(last_channel)->setName((char*)trackName.c_str());
        
		//sequence->getTrack(trackID)->freeReservedMemory();
    }//next track

		
	// in Midi, channel 9 is drums. (Aria assumes one track per channel - this is not always the case though)
    sequence->getTrack(9)->graphics->setEditorMode(DRUM);
    
    // erase empty tracks
    for(int n=0; n<sequence->getTrackAmount(); n++)
	{
        if(sequence->getTrack(n)->getNoteAmount() == 0)
		{
			sequence->deleteTrack(n);
            n=0;
        }
    }
    
    sequence->setXScrollInPixels(0);
    sequence->setFirstMeasure(0);
    
    // set song length
	std::cout << (int)ceil( (float)lastEventTick/(float)sequence->ticksPerMeasure() ) << " measures" << std::endl;
    sequence->mainFrame->changeMeasureAmount( (int)ceil( (float)lastEventTick/(float)sequence->ticksPerMeasure() ) );
    sequence->setMeasureAmount( (int)ceil( (float)lastEventTick/(float)sequence->ticksPerMeasure() ) );
    
    sequence->mainFrame->updateTopBarForSequence(sequence);
    sequence->setZoom(100);
	
    sequence->importing = false;
    
    return true;
    
}

}
