/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "Config.h"
#include "Clipboard.h"

#include "Midi/Track.h"
#include "Midi/Sequence.h"
#include "Midi/ControllerEvent.h"

#include "GUI/GraphicalTrack.h"
#include "GUI/MainFrame.h"
#include "GUI/GLPane.h"
#include "GUI/MeasureBar.h"

#include "Pickers/MagneticGrid.h"
#include "Pickers/InstrumentChoice.h"
#include "Pickers/DrumChoice.h"

#include "Editors/KeyboardEditor.h"
#include "Editors/GuitarEditor.h"
#include "Editors/RelativeXCoord.h"
#include "Editors/DrumEditor.h"
#include "Editors/ControllerEditor.h"

#include "IO/IOUtils.h"

#include <iostream>

#include "wx/utils.h"

namespace AriaMaestosa {

Track::Track(MainFrame* parent, Sequence* sequence)
{
    INIT_LEAK_CHECK();
	
	name = (char*)wxString(
						   wxGetTranslation("Untitled")
						   ).c_str();
	
    Track::frame=parent;
    
    glPane = parent->glPane;
    Track::sequence = sequence;
    graphics = new GraphicalTrack(this, sequence);

    graphics->createEditors();

}

Track::~Track(){
    
#ifdef _MORE_DEBUG_CHECKS
    //std::cout << "~Track BEGIN" << std::endl;
#endif
    
    delete graphics;
    notes.clearAndDeleteAll();
	noteOff.clearWithoutDeleting();
    controllerEvents.clearAndDeleteAll();
    
#ifdef _MORE_DEBUG_CHECKS
    //std::cout << "~Track END" << std::endl;
#endif
    
}

/*
 * Adds a note to the track. Returns true if note was added, returns false if it was rejected.
 */
bool Track::addNote(const int pitchID, const int startTick, const int endTick, const int volume, const int string){
    
	// reject empty/malformed notes
    if(startTick == endTick or startTick>endTick or startTick<0)
	{
        selectNote(ALL_NOTES, false); 
        return false;
    }
	
	// reject notes on strings that don't exist, i.e. the user clicked above or below the guitar)
    if(string!= -1)
	{
        if(string<0 or (unsigned int)string>graphics->guitarEditor->tuning.size()-1)
		{
            selectNote(ALL_NOTES, false); 
            return false;
        }
    }
    
    saveUndoMemory();
    
    Note* tmp_note;
    if(string==-1) tmp_note = new Note(graphics, pitchID, startTick, endTick, volume);
    else tmp_note = new Note(graphics, pitchID, startTick, endTick, volume, string, 0);
    
    const bool success = addNote( tmp_note );
    
	if(success)
	{
		// select last added note
		selectNote(ALL_NOTES, false, true); // select last added note
		tmp_note->setSelected(true);
    }
	
	tmp_note->play(true);
	
    return success;
}

bool Track::addNote(Note* note, bool check_for_overlapping_notes)
{

	bool noteAdded=false;
	
	//------------------------ place note on -----------------------
    // iterate through notes to place them in order
    const int noteAmount = notes.size();
    for(int n=0; n<noteAmount; n++)
	{
		// check for overlapping notes
		// the only time where this is not checked is when pasting, because it is then logical that notes are pasted on top of their originals
		if(!sequence->importing and check_for_overlapping_notes and notes[n].startTick == note->startTick and notes[n].pitchID == note->pitchID and
		   (graphics->editorMode!=GUITAR or notes[n].getString()==note->getString()) /*in guitar mode string must also match to be considered overlapping*/ )
		{
			std::cout << "overlapping notes: rejected" << std::endl;
			return false;	
		}
		
        if(notes[n].startTick > note->startTick)
		{
            
            notes.add(note, n);
			noteAdded=true;
            break;
        }//endif
		
		
    }//next
	
    if(!noteAdded)
	{
		// add note at the end, or if the track does not yet contain any note
		// (both cases in which the for loop either won't find a location for the note, or won't iterate at all)
		notes.push_back(note);
	}
	
	//------------------------ place note off -----------------------
    // iterate through notes to place them in order
	
	noteAdded = false;
	
    const int noteOffAmount = noteOff.size();
    for(int n=0; n<noteOffAmount; n++)
	{
        if(noteOff[n].endTick > note->endTick)
		{
            noteOff.add(note, n);
			noteAdded = true;
            break;
        }//endif
    }//next
    
	if(!noteAdded)
	{
		// add note at the end, or if the track does not yet contain any note
		// (both cases in which the for loop either won't find a location for the note, or won't iterate at all)
		noteOff.push_back(note);
	}
		
	
    return true;
	
}

int Track::getGridDivider(){
    return graphics->grid->divider;
}


// ------------------------------------- get information on notes -------------------------------------
// all IDs are relative to note one start (always in 'notes' vector, never in 'noteOff' vector)
// (i.e. if note starts second and ends fourth, you'll pass ID=2, not 4, even to get note end.)

int Track::getNoteStartInPixels(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    
    return (int)( notes[id].startTick * sequence->getZoom() );
}

int Track::getNoteEndInPixels(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());

    return (int)( notes[id].endTick * sequence->getZoom() );
}

int Track::getNoteStartInMidiTicks(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    
    return notes[id].startTick;
}

int Track::getNoteEndInMidiTicks(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    
    return notes[id].endTick;
}

int Track::getNotePitchID(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    return notes[id].pitchID;
}

bool Track::isNoteSelected(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    return notes[id].isSelected();
}

int Track::getNoteAmount()
{
    return notes.size();
}

int Track::getNoteVolume(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    return notes[id].volume;
}

int Track::getNoteString(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());
    
    //if(notes[id].tuning == NULL) notes[id].setTuning(&graphics->guitarEditor->tuning); // make sure the note knows the tuning
    if(notes[id].getString() == -1) notes[id].findStringAndFretFromNote();
    
    return notes[id].getString();
}

int Track::getNoteFret(const int id)
{
    assertExpr(id,>=,0);
    assertExpr(id,<,notes.size());

    //if(notes[id].tuning == NULL) notes[id].setTuning(&graphics->guitarEditor->tuning); // make sure the note knows the tuning
    if(notes[id].getFret() == -1) notes[id].findStringAndFretFromNote();
    
    return notes[id].getFret();
}

// returns the amount of ALL types of controller, not only of specified type
// the only goal of id is to determine whether the app is searching for a control event or for a tempo event
int Track::getControllerEventAmount(const int controllerTypeID)
{
    if(controllerTypeID==201 /*tempo*/) return sequence->tempoEvents.size();
    return controllerEvents.size();
}

ControllerEvent* Track::getControllerEvent(const int id, const int controllerTypeID)
{
    assertExpr(id,>=,0);
    if(controllerTypeID==201 /*tempo*/){
        assertExpr(id,<,sequence->tempoEvents.size());
    } else {
        assertExpr(id,<,controllerEvents.size());
    }
    
    if(controllerTypeID==201 /*tempo*/) return &sequence->tempoEvents[id];
    return &controllerEvents[id];
}

// ------------------------------------- modify notes -------------------------------------

/*
 * A midi controller, added from reading a file.
 * If we're reading the even from file, we can add it right away without further checks
 * because we know events won't overlap and are in time order.
 */

void Track::addControllerFromMidiFileReader(const int x, const int value, const int controller)
{
    controllerEvents.push_back(new ControllerEvent(sequence, controller, x, value) );
}

/*
 * The tick where the first note of the track starts playing.
 * Used mostly when scaling relative to track
 */

int Track::getFirstNoteTick(bool selectionOnly)
{

	if(!selectionOnly) return notes[0].startTick;
	
    const int noteAmount = notes.size();
    int tick = -1;
    
    for(int n=0; n<noteAmount; n++)
	{
		if( notes[n].isSelected() ) return notes[n].startTick;
		
        /*
        if(selectionOnly and !notes[n].isSelected()) continue;
        
        if(tick==-1 or notes[n].startTick<tick)
		{
            tick = notes[n].startTick;
        }//endif
        */
    }//next
    
    return tick;

}

/*
 * Add a single controller change event.
 */

void Track::addController(const int x, const int value, const int controller)
{
    
    assertExpr(controller,>=,0);
    assertExpr(controller,<,205);
    assertExpr(x,>=,0);
    assertExpr(value,>=,0);
    assertExpr(value,<,128);
    
    saveUndoMemory();
    
    // the vector events will be added to - allows to use the same code for multiple uses
    ptr_vector<ControllerEvent>* vector;
    
    // tempo events
    if(controller==201) vector = &sequence->tempoEvents;
    // controller and pitch bend events
    else vector = &controllerEvents;
        

    const int eventAmount=vector->size();
    for(int n=0; n<eventAmount; n++)
	{
        if( (*vector)[n].getTick()==x )
		{
            vector->erase(n);
            vector->add( new ControllerEvent(sequence, controller, x, value), n );
            return;
        }
        else if( (*vector)[n].getTick()>x )
		{
            vector->add( new ControllerEvent(sequence, controller, x, value), n );
            return;
        }//endif
        
    }//next
    
    vector->push_back(new ControllerEvent(sequence, controller, x, value) );
}

/*
 * Called when user drags mouse in the controller editor, adding a line (not a single events, a serie of events).
 * First, the method removes all events previously in this area.
 * Second, it adds the new events, being careful that the vector must stay in time order and that there must not be 2 consecutive events with the same value.
 */

void Track::addControllerSlide(const int x1, const int value1, const int x2, const int value2, const int controller)
{
    
    assertExpr(controller,>=,0);
    assertExpr(controller,<,205);
    assertExpr(x1,>=,0);
    assertExpr(x2,>=,0);
    assertExpr(value1,>=,0);
    assertExpr(value1,<,128);
    assertExpr(value2,>=,0);
    assertExpr(value2,<,128);

    saveUndoMemory();
        
    int addedAmount=0;
    
    // the vector events will be added to - allows to use the same code for multiple uses
    ptr_vector<ControllerEvent>* vector;
    
    // tempo events
    if(controller==201) vector = &sequence->tempoEvents;
    // controller and pitch bend events
    else vector = &controllerEvents;
    
    const int eventAmount=vector->size();
    
    
    // track is empty, events can be added without any further checking
    if(eventAmount==0)
	{
        
        int previous_value = value1;
        
        vector->add( new ControllerEvent(sequence, controller, x1, previous_value), 0 );
        addedAmount++;
        
        for(int tick=0; tick<x2-x1; tick++)
		{
            const int newvalue =
            (int)(
                  value1 + (value2-value1)*((float)tick / (float)(x2-x1))
                  );
            
            if(newvalue == previous_value )continue;
            
            vector->add( new ControllerEvent(sequence, controller, x1+tick, newvalue), addedAmount );
            addedAmount++;
            previous_value = newvalue;
        }//next
        return;
    }//end if
    
    // track is not empty; find which events need to be removed and where to add the new ones
    for(int n=0; n<eventAmount; n++)
	{
        
        if( (*vector)[n].getTick()>=x1 or n+1==eventAmount /*end of track*/ )
		{
            
			// erae all notes found in area where events need to be added
            int eraseID=n;
            while( eraseID<vector->size() and
                   (*vector)[eraseID].getTick()>=x1 and
                   (*vector)[eraseID].getTick()<=x2)
			{
#ifdef _MORE_DEBUG_CHECKS
				static int iters=0; iters++; if(iters>10000000){ std::cout << "infinite loop" << std::endl; exit(1);}
#endif
                if( (*vector)[eraseID].getController()==controller) vector->erase(eraseID);
                else eraseID++;
            }
            
            int previous_value = value1;
            
            // if we're at the end, events must not be added AT position (because nothing was deleted), but after position
            if(n==vector->size()-1) addedAmount++;

			// add first event
            vector->add( new ControllerEvent(sequence, controller, x1, previous_value), n+addedAmount );
            addedAmount++;
            
            // iterate from first tick to last tick of the slide, adding events all along, in time order
            for(int tick=0; tick<x2-x1; tick++)
			{
                // calculate next value in "slide", at the current tick
                const int newvalue =
                (int)(
                      value1 + (value2-value1)*((float)tick / (float)(x2-x1))
                      );
                
                // don't add useless controller events (two consecutive events with same value)
                if(newvalue == previous_value ) continue;
                
                // skip all events that happen before (i.e smaller tick) than the one that should be added
                // making sure that all controller events are placed in order
                while(n+addedAmount<vector->size()-1 and (*vector)[n+addedAmount].getTick() < x1+tick)
				{
                    n++;
                }
                
                vector->add( new ControllerEvent(sequence, controller, x1+tick, newvalue), n+addedAmount );
                addedAmount++;
                previous_value = newvalue;
            }//next
            
            return;
        }//endif
        
    }//next

}

/*
 * Called when user selects 'scale' from the menus and then clicks OK. This method takes care of actually scaling the notes.
 */

void Track::scale(float factor, int relative_to, bool selectionOnly){
    
    const int noteAmount=notes.size();

    for(int n=0; n<noteAmount; n++)
	{
        
        if(selectionOnly and !notes[n].isSelected()) continue; // skip unselected notes if we only want to affect selection
        
        notes[n].startTick = (int)(
                                   (notes[n].startTick-relative_to)*factor + relative_to
                                   );
        notes[n].endTick = (int)(
                                 (notes[n].endTick-relative_to)*factor + relative_to
                                 );
        
    }//next
	
	reorderNoteVector();
	reorderNoteOffVector();
    
}

void Track::setId(const int id){
    trackid = id;
}

/*
 * Make sure all notes in note on vector are in time order.
 */
void Track::reorderNoteVector()
{
	
	const int noteAmount = notes.size();
	
	// loop until nothing changed in a full iteration (meaning everything is in order)
	bool somethingChanged=true;
	while(somethingChanged)
	{
		somethingChanged=false;
		
#ifdef _MORE_DEBUG_CHECKS
		static int iters=0; iters++; if(iters>10000000){ std::cout << "infinite loop" << std::endl; exit(1);}
#endif
		
		for(int n=0; n<noteAmount-1; n++)
		{
			
			assertExpr(n+1,<,notes.size());
			//std::cout << n << " < " << notes.size() << std::endl;
			//std::cout << "n: " << notes[n].startTick << std::endl;
			//std::cout << "n+1: " << notes[n+1].startTick << std::endl;
			
			if(notes[n].startTick > notes[n+1].startTick)
			{
				notes.swap(n, n+1);
				if(n>2) n-= 2;
				somethingChanged = true;
			}
		}//next
	}//wend
	
}

/*
 * Make sure all notes in note off vector are in time order.
 */
void Track::reorderNoteOffVector(){
	
	const int noteAmount = noteOff.size();
	
	bool somethingChanged=true;
	while(somethingChanged){ // loop until nothing changed in a full iteration (meaning everything is in order)
		somethingChanged=false;
		
#ifdef _MORE_DEBUG_CHECKS
		static int iters=0; iters++; if(iters>10000000){ std::cout << "infinite loop" << std::endl; exit(1);}
#endif
		
		for(int n=0; n<noteAmount-1; n++)
		{
			
			if(noteOff[n].endTick > noteOff[n+1].endTick)
			{
				noteOff.swap(n, n+1);
				if(n>2) n-= 2;
				somethingChanged = true;
			}//end if
		}//next
	}//wend
	
}

/*
 * Erases all current data, replacing it by data previously stored as backup data.
 * This happens when user selects 'undo'.
 */
void Track::loadUndoMemory()
{
	
	if(!sequence->somethingDoneToUndo) return;
	
    const int noteAmount=sequence->notesUndoMemory.size();
    const int controllerEventsAmount=sequence->controllerUndoMemory.size();
	
    notes.clearAndDeleteAll();
	noteOff.clearWithoutDeleting();
    controllerEvents.clearAndDeleteAll();
 
    for(int n=0; n<noteAmount; n++)
	{
		Note* tmp = new Note( sequence->notesUndoMemory[n] );
        notes.push_back( tmp );
		noteOff.push_back( tmp );
    }//next
	
	reorderNoteOffVector();
	
    for(int n=0; n<controllerEventsAmount; n++)
	{
        controllerEvents.push_back( new ControllerEvent( sequence->controllerUndoMemory[n] ) );
    }//next
    
    sequence->somethingDoneToUndo = false;
}

/*
 * Saves current state just before doing something that will change notes/controllers.
 * This can be used laters to undo.
 */

void Track::saveUndoMemory(){
    if(sequence->importing) return;
    
	sequence->somethingDoneToUndo = true;
		
    sequence->lastModifiedTrack=trackid;
    sequence->notesUndoMemory.clearAndDeleteAll();
    sequence->controllerUndoMemory.clearAndDeleteAll();
    
    const int noteAmount=notes.size();
	
    for(int n=0; n<noteAmount; n++)
	{
        sequence->notesUndoMemory.push_back( new Note( notes[n] ) );
    }//next
	
    const int controllerEventsAmount=controllerEvents.size();
	
    for(int n=0; n<controllerEventsAmount; n++)
	{
        sequence->controllerUndoMemory.push_back( new ControllerEvent( controllerEvents[n] ) );
    }//next
    
}

void Track::selectNote(const int id, const bool selected, bool ignoreModifiers)
{
    
    assert(id != SELECTED_NOTES); // not supported in this function
    
    
    if(!glPane->isSelectMorePressed() and !glPane->isSelectLessPressed()) ignoreModifiers=true; // if no modifier is pressed, don't do any special checks
    
    // ----------------------------- select/deselect all notes ----------------------------
    if(id==ALL_NOTES)
	{
		// if this is a 'select none' command, unselect any selected measures in the top bar
		if(selected == FALSE) glPane->measureBar->unselect();
		
        if(graphics->editorMode == CONTROLLER)
		{ // controller editor must be handled differently
            graphics->controllerEditor->selectAll( selected );
        }
		else
		{
            
            if(ignoreModifiers)
			{
                for(int n=0; n<notes.size(); n++)
				{
                    notes[n].setSelected(selected);   
                }//next
            }//end if
            
        }// end if
        
    }
	else  // ----------------------------- select/deselect one specific note ----------------------------
	{
		
        assertExpr(id,>=,0);
        assertExpr(id,<,notes.size());

        // if we ignore +/- key modifiers, just set the value right away
        if(ignoreModifiers) notes[id].setSelected(selected);
        else
		{ // otherwise, check key modifiers and set value accordingly
            if(selected)
			{
                if(glPane->isSelectMorePressed()) notes[id].setSelected(true);
                else if(glPane->isSelectLessPressed()) notes[id].setSelected( !selected );
            }
        }//end if
		
    }//end if
}

void Track::moveNotes(int relativeX, int relativeY, int noteID){


    assert(noteID != ALL_NOTES); // not supported in this function (mostly bacause not needed, but could logically be implmented)
    
    saveUndoMemory();
    
    if(noteID==SELECTED_NOTES)
	{
        
		bool played = false;
		
        for(int n=0; n<notes.size(); n++)
		{
            if(!notes[n].isSelected()) continue;
            
            notes[n].move(relativeX, relativeY, graphics->editorMode);
			
			if(!played)
			{
				if(relativeY != 0) notes[n].play(true);
				else notes[n].play(false);
				played = true;
			}
        }//next
		
    }
	else
	{ 
		// move a single note
        assertExpr(noteID,>=,0);
        assertExpr(noteID,<,notes.size());
        
        notes[noteID].move(relativeX, relativeY, graphics->editorMode);
		
		if(relativeX != 0)
		{
			if(relativeY != 0) notes[noteID].play(true);
			else notes[noteID].play(false);
		}
    }
	
    reorderNoteVector();
	reorderNoteOffVector();
}

void Track::prepareNotesForGuitarEditor(){
	
    for(int n=0; n<notes.size(); n++)
	{
        notes[n].checkIfStringAndFretMatchNote(true);
    }
	
}

void Track::playNote(const int id, const bool noteChange){
	assertExpr(id,<,notes.size());
	assertExpr(id,>=,0);
	
	notes[id].play(noteChange);
	
}

void Track::setNoteVolume(int volume, int id){

    saveUndoMemory();
    
    assertExpr(id,<,notes.size());
    
    assert(id != ALL_NOTES); // not supported in this function (mostly bacause not needed, but could logically be implmented)
    
    if(id == SELECTED_NOTES)
	{
        
		bool played = false;
        for(int n=0; n<notes.size(); n++)
		{
            if(notes[n].isSelected())
			{
				notes[n].setVolume( volume );
				if(!played)
				{
					notes[n].play(true);
					played = true;
				}
			}
        }//next note
		
    }
	else
	{
		// single note
		
        assertExpr(id,>=,0);
        assertExpr(id,<,notes.size());
        
        // if user changed the volume of a note that is not selected, change the volume of this note only
        notes[id].setVolume( volume );
		notes[id].play(true);
    }
    
}
void Track::setNoteEnd(const int tick, const int noteID){

    assert(noteID != ALL_NOTES); // not supported in this function (mostly bacause not needed, but could logically be implmented)
    assert(noteID != SELECTED_NOTES); // not supported in this function (mostly bacause not needed, but could logically be implmented)
 
    assertExpr(noteID,<,notes.size());
    assertExpr(noteID,>=,0);
    
    notes[noteID].setEnd(tick);
	
	reorderNoteOffVector();
}

void Track::resizeNotes(int relativeWidth, int noteID){
  
    saveUndoMemory();
    
    if(graphics->editorMode == DRUM) return;
    
    assert(noteID != ALL_NOTES); // not supported in this function (mostly bacause not needed, but could logically be implmented)
    
    if(noteID==SELECTED_NOTES)
	{
        
		bool played=false;
        for(int n=0; n<notes.size(); n++)
		{
            if(!notes[n].isSelected()) continue;
            
            notes[n].resize(relativeWidth);
			
			if(!played)
			{
				notes[n].play(false);
				played = true;
			}
        }//next
        
    }
	else
	{ 
        
        assertExpr(noteID,<,notes.size());
        assertExpr(noteID,>=,0);
        
        notes[noteID].resize(relativeWidth);
		notes[noteID].play(false);
    }
    
	reorderNoteOffVector();
}

/*
 * Deletes all selected notes
 */

void Track::deleteSelected()
{
    
    saveUndoMemory();
    
    if(graphics->editorMode == CONTROLLER)
	{
        
        const int from = graphics->controllerEditor->getSelectionBegin();
        const int to = graphics->controllerEditor->getSelectionEnd();
        const int type = graphics->controllerEditor->getCurrentControllerType();
		
		if(type != 201 /*tempo*/)
		{
			// remove controller events
			for(int n=0; n<controllerEvents.size(); n++)
			{
				
				if(controllerEvents[n].getController() != type) continue; // in another controller
				   
				   const int tick = controllerEvents[n].getTick();
				   
				   if(tick<from or tick>to) continue; // this event is not concerned by selection
				   
				   controllerEvents.erase(n);
				   n--;
			}//next

		}
        else
        {
			// remove tempo events
			for(int n=0; n<sequence->tempoEvents.size(); n++)
			{
				
				const int tick = sequence->tempoEvents[n].getTick();
				
				if(tick<from or tick>to) continue; // this event is not concerned by selection
				
				sequence->tempoEvents.erase(n);
				n--;
			}//next
         }

    }
    else
    {
        
        for(int n=0; n<notes.size(); n++)
		{
            if(!notes[n].isSelected()) continue;
            
			// also delete corresponding note off event
			for(int i=0; i<noteOff.size(); i++)
			{
				if(&noteOff[i] == &notes[n])
				{
					noteOff.remove(i);
					break;
				}
			}
			
            notes.erase(n);
			
            n--;
        }//next
        
    }
       
	
	reorderNoteOffVector();
}

void Track::removeNote(const int id)
{
	
	// also delete corresponding note off event
	for(int i=0; i<noteOff.size(); i++)
	{
		if(&noteOff[i] == &notes[id])
		{
			noteOff.remove(i);
			break;
		}
	}
	
	notes.erase(id);
	
}

/*
 * In guitar editor, changes the number on a fret, also changing the note. This is called when user changes frets with ctrl + arrow keys.
 */

void Track::shiftFrets(const int amount, const int id)
{
    
    saveUndoMemory();
    
    if(graphics->editorMode != GUITAR) return;
    
    assert(id != ALL_NOTES); // not supported in this function (mostly bacause not needed, but could logically be implmented)
    
    // only accept to do this in guitar mode
    if(graphics->editorMode != GUITAR)  return;
    
	// concerns all selected notes
    if(id==SELECTED_NOTES)
	{
        
		bool played = false;
        for(int n=0; n<notes.size(); n++)
		{
            if(!notes[n].isSelected()) continue;
            
            notes[n].shiftFret(amount);
			
			if(!played)
			{
				notes[n].play(true);
				played = true;
			}
        }//next
        
    }
	// only concernes one specific note (FIXME - is this used at all?)
	else
	{ 
        assertExpr(id,>=,0);
        assertExpr(id,<,notes.size());
        
        notes[id].shiftFret(amount);
		notes[id].play(true);
    }
    
}


void Track::numberPressed(int number)
{
    
    saveUndoMemory();
    
    if(graphics->editorMode != GUITAR) return;
       
	bool played = false;
    for(int n=0; n<notes.size(); n++)
	{
        if(!notes[n].isSelected()) continue;
        
        notes[n].setFret(number);
		if(!played)
		{
			notes[n].play(true);
			played = true;
		}
        return;
    }//next
	
    
}

void Track::updateNoteTuning(const int id)
{
    assert(id != SELECTED_NOTES); // not supported in this function
    
    if(id==ALL_NOTES)
	{
        
        for(int n=0; n<notes.size(); n++)
		{
            notes[n].checkIfStringAndFretMatchNote(true);
        }//next
        
    }
	else
	{
        assertExpr(id,>=,0);
        assertExpr(id,<,notes.size());
        
        notes[id].checkIfStringAndFretMatchNote(true);
    }
}

/*
 * Keep same note, place on a different fret.
 */

void Track::shiftString(const int amount, const int id)
{
    
    saveUndoMemory();
    
    if(graphics->editorMode != GUITAR) return;
    
    assert(id != ALL_NOTES); // not supported in this function (mostly because not needed, but could logically be implmented)
    
    // only accept to do this in guitar mode
    if(graphics->editorMode != GUITAR)
        return;
    
    if(id==SELECTED_NOTES)
	{
        
		bool played=false;
        for(int n=0; n<notes.size(); n++)
		{
            if(!notes[n].isSelected()) continue;
            
            notes[n].shiftString(amount);
			
			if(!played)
			{
				notes[n].play(false);
				played = true;
			}
        }//next
        
    }
	else
	{  
        assertExpr(id,>=,0);
        assertExpr(id,<,notes.size());
        
        notes[id].shiftString(amount);
		notes[id].play(false);
    }
    
}

// ------------------------------------- midi playback / IO -------------------------------------

// returns smallest values, ignoring -1, a being prioritary to b (returns -1 for none, 0 for a, 1 for b)
int getActiveMin(int a, int b)
{
	if(a==-1 and b==-1) return -1;
	if(a==-1) return 1;//b;
	if(b==-1) return 0;//a;
	if(a <= b) return 0;//a;
	return 1;//b;
	
}

// returns smallest values, ignoring -1, a being prioritary to b, and b prioritary to c
// (returns -1 for none, 0 for a, 1 for b, 2 for c)
int getActiveMin(int a, int b, int c)
{
	
	const int result1 = getActiveMin(b,c);

	int passnumber;
	if(result1==0) passnumber = b;
	else if(result1==1) passnumber = c;
	else if(result1==-1) passnumber =-1;
	
	const int result2 = getActiveMin( a, passnumber );
	
	if(result2==0) return 0;
	else
	{
		if(result1==-1) return -1;
		
		//if(result1==0) return 1;
		//if(result1==1) return 2;
		return result1+1;
	}
	
}

void Track::reserveMemoryForEvents(int eventAmount)
{
	// tell the vectors to reserve some memory so that it's not constantly resized during import
	// not used ATM caused more problems and bringed little
	notes.contentsVector.reserve(eventAmount);
	noteOff.contentsVector.reserve(eventAmount);
}

/*
 * Add Midi Events to JDKMidi track object
 */

int Track::addMidiEvents(jdkmidi::MIDITrack* midiTrack, int channel, int firstMeasure, bool selectionOnly, int& startTick)
{
	
	// ignore track if it's muted
	// (but for some reason drum track can't be completely omitted)
	// if we only play selection, ignore mute and play anyway
    if(graphics->muted and graphics->editorMode != DRUM and !selectionOnly)
		return -1;
    
    if(graphics->editorMode == DRUM) channel=9;
    
    // when previewing selected notes (start by finding the note that plays first, to start playing at the right place and note from the beginning)
    int firstNoteStartTick = -1;
    int selectedNoteAmount=0;
	
    if(selectionOnly)
	{
		
        const int noteAmount=notes.size();
        for(int n=0; n<noteAmount; n++)
		{
            if(notes[n].isSelected())
			{
                if(notes[n].startTick < firstNoteStartTick or firstNoteStartTick==-1) firstNoteStartTick = notes[n].startTick;
                selectedNoteAmount++;
            }
        }
        
        if(firstNoteStartTick == -1) return -1; // error, no note was found.
        if(selectedNoteAmount == 0) return -1; // error, no note was found.
		
    }
	else
	{
        firstNoteStartTick = firstMeasure*sequence->ticksPerMeasure();
        startTick = firstNoteStartTick;
    }
    
    // set bank
    {
        jdkmidi::MIDITimedBigMessage m;
        
        m.SetTime( 0 );
        m.SetControlChange( channel, 0, 0 );
        
        if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl; exit(1); }
        
        m.SetTime( 0 );
        m.SetControlChange( channel, 32, 0 );
        
        if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl; exit(1); }
    }
    
    // set instrument
    {
        jdkmidi::MIDITimedBigMessage m;
        
        m.SetTime( 0 );

        if(graphics->editorMode == DRUM) m.SetProgramChange( channel, graphics->drumEditor->drumKit->getDrumID() );
        else m.SetProgramChange( channel, graphics->keyboardEditor->instrument->getInstrumentID() );
        
        if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding instrument at track beginning!" << std::endl; }
    }
    
    // set track name
    {

        jdkmidi::MIDITimedBigMessage m;
        
        m.SetText( 3 );
        m.SetByte1( 3 );
        
        jdkmidi::MIDISystemExclusive sysex( (unsigned char*)name.c_str(), name.size()+1, name.size()+1, false);
        
        m.CopySysEx( &sysex );

        m.SetTime( 0 );
        
        if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl; exit(1); }

        
    }
    
    // set maximum volume
    {
        jdkmidi::MIDITimedBigMessage m;
        
        m.SetTime( 0 );
        m.SetControlChange( channel, 7, 127 );
        
        if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl; exit(1); }
    }
    
	// ----------------------------------- add events in order --------------------------
	/*
	 * The way this section works:
	 *
	 * there are currently 3 possible source of events in a track (apart those above, like instrument and track name): note on, note off, controller change.
	 * each type of event is stored in its own vector, in time order.
	 * the variables below store the current event (i.e. the first event that hasn't yet been added)
	 * the section loops, and with it each iteration it checks the current tick of the 3 current events,
	 * then picks the one with the smallest tick and adds it to the track.
	 *
	 */
	
	int note_on_id = 0;
	int note_off_id = 0;
	int control_evt_id = 0;
	
	const int noteOnAmount = notes.size();
	const int noteOffAmount = noteOff.size();
	const int controllerAmount = controllerEvents.size();
	
	// find track end
    int last_event_tick=0;
	
	// if muted and drums, return now
	if(graphics->muted and graphics->editorMode == DRUM and !selectionOnly) return -1;
	
	while(true)
	{
		
#ifdef _MORE_DEBUG_CHECKS
		static int iters=0; iters++; if(iters>10000000){ std::cout << "infinite loop" << std::endl; exit(1);}
#endif
		
		// if we only want to play what's selected, skip unselected notes
		if(selectionOnly)
		{
			while(note_on_id < noteOnAmount and !notes[note_on_id].isSelected()) note_on_id++;
			while(note_off_id < noteOffAmount and !noteOff[note_off_id].isSelected()) note_off_id++;
		}
		
		const int tick_on      = (note_on_id < noteOnAmount)     ?   notes[note_on_id].startTick - firstNoteStartTick      :  -1;
		const int tick_off     = (note_off_id < noteOffAmount)   ?   noteOff[note_off_id].endTick - firstNoteStartTick     :  -1;
		
		// ignore control events when only playing selection
		const int tick_control = (control_evt_id < controllerAmount and !selectionOnly)  ?   controllerEvents[control_evt_id].getTick() - firstNoteStartTick : -1;
		
		const int activeMin = getActiveMin( tick_off, tick_control, tick_on );
		
		if(activeMin==-1) break; // all events have been added
		
		jdkmidi::MIDITimedBigMessage m;
		
		//  ------------------------ add note on event ------------------------ 
		if(activeMin == 2)
		{

            const int time=notes[note_on_id].startTick - firstNoteStartTick;
            if(!(time < 0))
			{
				
				m.SetTime( time );
				
				
				if(graphics->editorMode == DRUM) m.SetNoteOn( channel, notes[note_on_id].pitchID, notes[note_on_id].volume );
				else m.SetNoteOn( channel, 131-notes[note_on_id].pitchID, notes[note_on_id].volume );
				
				// find track end
				if(notes[note_on_id].endTick > last_event_tick) last_event_tick = notes[note_on_id].endTick;
				
				if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
			}
			
			note_on_id++;
		}
		//  ------------------------ add note off event ------------------------ 
		else if(activeMin == 0)
		{
			
			const int time=noteOff[note_off_id].endTick - firstNoteStartTick;
            if(!(time < 0))
			{
				
				m.SetTime( time );

				if(graphics->editorMode == DRUM) m.SetNoteOff( channel, noteOff[note_off_id].pitchID, 0 );
				else m.SetNoteOff( channel, 131-noteOff[note_off_id].pitchID, 0 );
				
				// find track end
				if(time > last_event_tick) last_event_tick = time;
				
				if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
			}
			note_off_id++;
		}
		//  ------------------------ add control change event ------------------------ 
		else if(activeMin == 1)
		{
			
			const int controllerID = controllerEvents[control_evt_id].getController();
            
            if(controllerID==200)
			{ // pitch bend
                
                const int time=controllerEvents[control_evt_id].getTick() - firstNoteStartTick;
                if(!(time < 0))
				{
					
					m.SetTime( time );

					int pitchBendVal = (127-controllerEvents[control_evt_id].getValue())*128-128*128;
					
					if(pitchBendVal > 0) pitchBendVal-=8192;
					else if(pitchBendVal < 0) pitchBendVal +=8192;
					
					m.SetPitchBend( channel,
									pitchBendVal );
									
					if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
				}
			}
			// other controller
			else
			{
				
				const int time=controllerEvents[control_evt_id].getTick() - firstNoteStartTick;
				if(!(time < 0))
				{
					
					m.SetTime( time );

					m.SetControlChange( channel,
										controllerID,
										127-controllerEvents[control_evt_id].getValue() );
										
					if( !midiTrack->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
				}
			}//endif
			
			control_evt_id++;
			
		}//end if
		
	}//wend
	
	
	if(selectionOnly) startTick = firstNoteStartTick;
    return last_event_tick - firstNoteStartTick;
	
	
}

/*
 * Insert empty measures in the middle of a song
 */

void Track::insertEmptyMeasures(int measureID, int amount)
{
	// convert measures into midi ticks
	const int amountInTicks = amount * sequence->ticksPerMeasure();
	const int afterTick = measureID * sequence->ticksPerMeasure() - 1;
	
	// move all notes that are after given start tick by the necessary amount
	const int noteAmount = notes.size();
	for(int n=0; n<noteAmount; n++)
	{
        if(notes[n].startTick > afterTick)
        {
			notes[n].startTick += amountInTicks;
			notes[n].endTick += amountInTicks;
		}
    }
	const int controlAmount = controllerEvents.size();
	for(int n=0; n<controlAmount; n++)
	{
        if(controllerEvents[n].getTick() > afterTick)
        {
			controllerEvents[n].setTick( controllerEvents[n].getTick() + amountInTicks );
		}
    }
	
}

void Track::removeMeasures(int from_measure, int to_measure)
{
	// find the amount of ticks that will be removed. This will be used to move back notes located after the area that is removed.
	const int amountInTicks = (to_measure-from_measure) * sequence->ticksPerMeasure();
	
	// find the range of ticks that need to be removed (convert measure IDs to midi ticks)
	const int from_tick = from_measure * sequence->ticksPerMeasure() - 1;
	const int to_tick = to_measure *  + sequence->ticksPerMeasure();
	
	// move all notes that are after given start tick by the necessary amount
	for(int n=0; n<notes.size(); n++)
	{
        if(notes[n].startTick > from_tick and notes[n].startTick < to_tick)
        {
			removeNote(n);
			n--;
		}
		else if(notes[n].startTick >= to_tick)
        {
			notes[n].startTick -= amountInTicks;
			notes[n].endTick -= amountInTicks;
		}
    }
	
	for(int n=0; n<controllerEvents.size(); n++)
	{
		// delete all controller events located in the area to be deleted
        if(controllerEvents[n].getTick() > from_tick and controllerEvents[n].getTick() < to_tick)
        {
			controllerEvents.markToBeDeleted(n);
		}
		// move all controller events that are after given start tick by the necessary amount
		else if(controllerEvents[n].getTick() >= to_tick)
        {
			controllerEvents[n].setTick( controllerEvents[n].getTick() - amountInTicks );
		}
    }
	controllerEvents.removeMarked();
	
	reorderNoteVector();
	reorderNoteOffVector();
}

void Track::snapNotesToGrid()
{
    
    for(int n=0; n<notes.size(); n++)
	{
        if(!notes[n].isSelected()) continue;
        
        notes[n].startTick = graphics->getCurrentEditor()->snapMidiTickToGrid( notes[n].startTick );
        notes[n].endTick = graphics->getCurrentEditor()->snapMidiTickToGrid( notes[n].endTick );
    }
    
}

/*
 * Midi forbids that two notes with the same pitch play at the same time on the same channel.
 * This method removes some of these duplicates (only those that start at the same time)
 */

void Track::removeOverlapping()
{
	wxBeginBusyCursor();
	
	const int noteOffAmount = noteOff.size();
	const int noteAmount = notes.size();
	
	// compare all notes to see if they match
	for(int n1=0; n1<noteAmount; n1++)
	{
		
		for(int n2=0; n2<noteAmount; n2++)
		{
			
			if(n1 == n2) continue; // don't compare a note with itself
			if(notes.isMarked(n1)) continue; // skip notes already removed
			if(notes.isMarked(n2)) continue; // skip notes already removed
			
			// both notes have the same pitch, they are candidates for overlapping
			if(notes[n1].pitchID == notes[n2].pitchID)
			{
				const int from = std::min(notes[n1].startTick, notes[n2].startTick);
				const int to = std::max(notes[n1].endTick, notes[n2].endTick);
				// total length of both notes
				const int length =  (notes[n1].endTick - notes[n1].startTick) + (notes[n2].endTick - notes[n2].startTick);
				
				// if difference between beginning of first note and ending of second note is smaller than their lengths, then the notes overlap
				// if they both begin at the same tick they also overlap
				if( (to - from < length) or (to - from == 0) )
				{

					// find the correspondant note off event and mark it to be removed too
					for(int noff=0; noff<noteOffAmount; noff++)
					{
						if( &notes[n1] == &noteOff[noff] )
						{
							noteOff.markToBeRemoved(noff);
							break;
						}
					}
					
					notes.markToBeDeleted(n1);
				}
			
			
			}// end if pitches match
			
		}//next n2
		
	}//next n1
	
	notes.removeMarked();
	noteOff.removeMarked();
	wxEndBusyCursor();
	
	sequence->glPane->render();
	
}

void Track::copy()
{
    
    if(graphics->editorMode == CONTROLLER)
	{
		wxBell();
		return; // no copy/paste in controller mode
    }
	
	Clipboard::clear();
    
	int beginning=-1;
    // place all selected notes into clipboard
    for(int n=0; n<notes.size(); n++)
	{
        if(!notes[n].isSelected()) continue;
        
        Note* tmp=new Note(notes[n]);
		Clipboard::add(tmp);
		
		// if in guitar mode, make sure string/fret and note match
		if(graphics->editorMode == GUITAR) Clipboard::getNote( Clipboard::getSize()-1 )->checkIfStringAndFretMatchNote(false);
		
		// find beginning of the first note
		if(notes[n].startTick < beginning or beginning==-1) beginning=notes[n].startTick;
		
    }//next
    
    // remove all empty measures before notes, so that they appear in the current measure when pasting
    const int lastMeasureStart = (beginning-beginning%sequence->ticksPerMeasure()); // we want to copy notes relative to the first measure copied notes are in
    
	const int clipboard_size = Clipboard::getSize();
	for(int n=0; n<clipboard_size; n++)
	{
        Clipboard::getNote(n)->move( -lastMeasureStart, 0, graphics->editorMode);
    }
	
	sequence->notes_shift_when_no_scrolling = lastMeasureStart;
}


/*
 * If atMouse is true, the first note in the clipboard is placed at the mouse cursor.
 * Otherwise, they are pasted in the first visible measure, still keeping the same position within the measure they had when being copied
 */

void Track::paste(bool atMouse)
{
    
	// FIXME - clean up, remove goto, etc.
	
    if(graphics->editorMode == CONTROLLER)
	{
		wxBell();
		return; // no copy/paste in controller mode
	}
    if(Clipboard::getSize() == 0) return; // nothing copied
    
    saveUndoMemory();
    
	
    // used to find when the first note is played.
    // If we're pasting at the mouse cursor, it is necessary to know that so that the first note is next to the cursor.
    // However, if we're pasting in the first measure, we don't want this info since we want the notes to keep their location within the measure.
    int beginning=-1;
    if(!atMouse) beginning=0;
    
    // unselected previously selected notes
    for(int n=0; n<notes.size(); n++) notes[n].setSelected(false);
    
    // find where notes begin if necessary
    if(atMouse)
		beginning = Clipboard::getNote(0)->startTick;
    
    int shift=0;
    
    // if "paste at mouse", use its location to know where to paste notes
    wxPoint mouseLoc=wxGetMousePosition();
    wxPoint trackMouseLoc=glPane->ScreenToClient(mouseLoc);
    if( atMouse and
        trackMouseLoc.y > graphics->getCurrentEditor()->getEditorYStart() and
        trackMouseLoc.y < graphics->getCurrentEditor()->getYEnd() and
        trackMouseLoc.x > getEditorXStart())
	{
        
        RelativeXCoord<int> mx(sequence);
		mx.setValue(trackMouseLoc.x, WINDOW);
        
        shift=graphics->getCurrentEditor()->snapMidiTickToGrid( sequence->getXScrollInMidiTicks() + mx.getRelativeTo(MIDI) );

	}
	else if(!atMouse and sequence->x_scroll_upon_copying == sequence->getXScrollInPixels() )
	{
		
		/*
		 * scrolling has not changed since copy (i.e. user probably hit 'copy' and 'paste' consecutively)
		 * in this case, we want the notes to be pasted exactly where they were copied
		 * so we just overwrite the shift variable to make notes paste where they were copied.
		 * after this, check if all notes will be visible when pasting like this. If not, fall back to regular pasting method
		 * that will take care of finding notes an appropriate location to be pasted onto.
		 */
		shift = sequence->notes_shift_when_no_scrolling;
		
		// find if first note will be visible in the location just calculated,
		// otherwise just go to regular pasting code, it will paste them within visible measures
		Note* tmp=new Note( *(Clipboard::getNote(0)) );
		
		// before visible area
		if((tmp->startTick+tmp->endTick)/2 + shift < sequence->getXScrollInMidiTicks())
			goto regular_paste;
		
		// after visible area
		RelativeXCoord<int> screen_width( sequence );
		screen_width.setValue( glPane->getWidth(), WINDOW  );
		
		if((tmp->startTick+tmp->endTick)/2 + shift > sequence->getXScrollInMidiTicks() + screen_width.getRelativeTo(MIDI) )
			goto regular_paste;
		
		
    }
	else
	{
		
regular_paste:
		/*
		 * This part will change to value of 'shift' variable to make sure pasted notes will be visible
		 */

        // if not "paste at mouse", find first visible measure
        const int lastMeasureStart = (sequence->getXScrollInMidiTicks()-sequence->getXScrollInMidiTicks()%sequence->ticksPerMeasure());
        shift=graphics->keyboardEditor->snapMidiTickToGrid(lastMeasureStart);
        
		// find if all notes will be visible in the location just calculated,
		// otherwise move them one more measure ahead (if measure is half-visible because of scrolling)
		const int clipboardSize = Clipboard::getSize();
		for(int n=0; n<clipboardSize; n++)
		{
			Note& tmp = *(Clipboard::getNote(n));
			
			//std::cout << (tmp.startTick + tmp.endTick)/2 + shift << " " << sequence->getXScrollInMidiTicks() << std::endl;
			
			// check if note is before visible area
			if((tmp.startTick + tmp.endTick)/2 + shift < sequence->getXScrollInMidiTicks())
			{
				shift+=sequence->ticksPerMeasure();
				break;
			}//end if
			
			// check if note is after visible area
			RelativeXCoord<int> screen_width( sequence );
			screen_width.setValue( glPane->getWidth(), WINDOW  );
			
			if((tmp.startTick + tmp.endTick)/2 > sequence->getXScrollInMidiTicks() + screen_width.getRelativeTo(MIDI) )
			{
				shift-=sequence->ticksPerMeasure();
				break;
			}//end if
			
		}//next
		
    }
    
    // add new notes
	const int clipboardSize = Clipboard::getSize();
    for(int n=0; n<clipboardSize; n++)
	{
        
        Note* tmp=new Note( *(Clipboard::getNote(n)) );
        
        tmp->move( shift , 0, graphics->editorMode);
        if(atMouse) tmp->move( -beginning , 0, graphics->editorMode);
        
        tmp->setParent( graphics );
        tmp->setSelected(true);
        
        if(graphics->editorMode == GUITAR) tmp->checkIfStringAndFretMatchNote(true);
        
        addNote( tmp, false );
    }//next
    
}

void Track::setName(char* name) {		Track::name = name;		}
std::string& Track::getName()   {		return name;			}

// -----------------------------------------------------------------------------------------------------
// ------------------------------------------- SAVE/LOAD -----------------------------------------------
// -----------------------------------------------------------------------------------------------------

void Track::saveToFile(wxFileOutputStream& fileout)
{
	
	writeData(wxString("\n<track name=\"") + wxString((char*)name.c_str()) + wxString("\">\n"), fileout );
	
	graphics->saveToFile(fileout);
	
	// notes
    for(int n=0; n<notes.size(); n++)
        notes[n].saveToFile(fileout);   
	
	// controller changes
    for(int n=0; n<controllerEvents.size(); n++)
        controllerEvents[n].saveToFile(fileout);   
	
	writeData(wxString("</track>\n\n"), fileout );
	

}

bool Track::readFromFile(irr::io::IrrXMLReader* xml)
{

	notes.clearAndDeleteAll();
	noteOff.clearWithoutDeleting(); // have already been deleted by previous command
    controllerEvents.clearAndDeleteAll();
	
	// parse XML file
	do
    {
        
        switch(xml->getNodeType())
        {
            case irr::io::EXN_TEXT:
                break;
            case irr::io::EXN_ELEMENT:
            {
				
                if (!strcmp("track", xml->getNodeName()))
				{

                    const char* name = xml->getAttributeValue("name");
					if(name!=NULL) setName( (char*)name );
					else
					{
						std::cout << "Missing info from file: track name" << std::endl;
						setName( "Untitled" );
					}
						

                }
                else if (!strcmp("editor", xml->getNodeName()))
				{
					if(! graphics->readFromFile(xml) )
						return false;
                }  
				else if (!strcmp("magneticgrid", xml->getNodeName()))
				{
					if(! graphics->readFromFile(xml) )
						return false;
                } 
				else if (!strcmp("instrument", xml->getNodeName()))
				{
					if(! graphics->keyboardEditor->instrument->readFromFile(xml) )
						return false;
                } 
				else if (!strcmp("drumkit", xml->getNodeName()))
				{
					if(! graphics->drumEditor->drumKit->readFromFile(xml) )
						return false;
                } 
				else if (!strcmp("guitartuning", xml->getNodeName()))
				{
					if(! graphics->readFromFile(xml) )
						return false;
                } 
				else if (!strcmp("note", xml->getNodeName()))
				{
					Note* temp = new Note(graphics);
					if(! temp->readFromFile(xml) )
					{
						std::cout << "A note was discarded because it is invalid" << std::endl;
						delete temp;
					}
					else
					{
						addNote( temp );
					}
                } 
				else if (!strcmp("controlevent", xml->getNodeName()))
				{
					
					ControllerEvent* temp = new ControllerEvent(sequence, 0, 0, 0);
					if( !temp->readFromFile(xml) )
					{
						// tehre was an error when trying to load control event
						std::cout << "A controller event was discarded because it is invalid" << std::endl;
						delete temp;
						//controllerEvents.push_back( temp );
					}
					else
					{
						// controller successfully loaded, add to controllers vector
						controllerEvents.push_back( temp );
						//delete temp;	
					}
                } 
				
			}// end case
				
                break;
            case irr::io::EXN_ELEMENT_END:
            {
                
                if (!strcmp("track", xml->getNodeName()))
				{
                    return true;
                }
            }
                break;
               
			default:break;
        }//end switch
		
    }while(xml && xml->read());
	
	return true;
	
}

}
